@@ -1,6 +1,7 @@
 import json
 import time
 from datetime import datetime
+from functools import lru_cache
 from http import HTTPStatus
 from typing import List
 
@@ -13,7 +14,6 @@ from api_management_local.exception_api import (ApiTypeDisabledException,
 from api_management_local.indirect import InDirect
 from item_local.item import Item
 from logger_local.MetaLogger import MetaLogger
-from python_sdk_remote.utilities import our_get_env
 from star_local.exception_star import NotEnoughStarsForActivityException
 
 from .ChannelProviderConstants import (AWS_SES_EMAIL_MESSAGE_PROVIDER_ID,
@@ -21,24 +21,23 @@ from .ChannelProviderConstants import (AWS_SES_EMAIL_MESSAGE_PROVIDER_ID,
                                        INFORU_MESSAGE_PROVIDER_ID)
 from .CompoundMessage import CompoundMessage
 from .MessageChannels import MessageChannel
-from .MessageConstants import (DEFAULT_HEADERS, SMS_MESSAGE_LENGTH,
-                               object_message)
+from .MessageConstants import DEFAULT_HTTP_HEADER, SMS_MESSAGE_LENGTH, object_message
 from .MessageImportance import MessageImportance
 from .Recipient import Recipient
 
-MESSAGE_TEST_SENDER_PROFILE_ID = our_get_env("MESSAGE_TEST_SENDER_PROFILE_ID", raise_if_not_found=False)
-
 
+# TODO recipients -> recipients_list (breaking change)
+# TODO: lines not tested - 83, 101, 107, 118-122, 128-139, 151, 162, 167-172, 177-209, 213-253, 259, 265-269, 275, 280, 292, 296, 299
 class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_message):
     """Message Local Class"""
 
-    def __init__(self, *, recipients: List[Recipient], api_type_id: int = None, campaign_id: int = None,
+    def __init__(self, *, recipients: List[Recipient] = None, api_type_id: int = None, campaign_id: int = None,
                  message_template_id: int = None,
                  original_body: str = None, original_subject: str = None,
                  importance: MessageImportance = MessageImportance.MEDIUM, message_id: int = None,
-                 is_http_api: bool = None, endpoint: str = None, is_debug: bool = False,
-                 headers: dict = DEFAULT_HEADERS, user_external_id: int = None, form_id: int = None,
-                 sender_profile_id: int = MESSAGE_TEST_SENDER_PROFILE_ID, is_test_data: bool = False,
+                 is_http_api: bool = None, endpoint_url: str = None, is_debug: bool = False,
+                 headers: dict = DEFAULT_HTTP_HEADER, user_external_id: int = None, form_id: int = None,
+                 sender_profile_id: int = None, is_test_data: bool = False,
                  is_require_moderator: bool = True) -> None:
         """message_dict contains a list of dicts, each with the following keys:
         ["sms_body_template", "email_subject_template", "email_body_html_template",
@@ -50,17 +49,19 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
                                  message_template_id=message_template_id, original_body=original_body, form_id=form_id,
                                  original_subject=original_subject, recipients=recipients, is_test_data=is_test_data,
                                  is_require_moderator=is_require_moderator)
+        if not self._recipients:
+            raise Exception(
+                "recipients parameter is required (couldn't be detected automatically based on the provided data)")
 
         # init instance variables
         self.importance = importance
         self._is_http_api = is_http_api
         self._api_type_id = api_type_id
-        self._endpoint = endpoint
+        self._endpoint_url = endpoint_url
         self._headers = headers
         self.__user_external_id = user_external_id
-        self.__recipients = recipients
         self._campaign_id = campaign_id
-        self.sender_profile_id = sender_profile_id
+        self.__sender_profile_id = sender_profile_id
 
         # init instances
         self.__indirect = InDirect()
@@ -68,11 +69,12 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
         self.api_management_manager = APIManagementsManager()
 
         # decide how to send (channel & provider) for each recipient
-        self.channel_ids_per_recipient = {recipient.get_profile_id(): self.get_msg_type(recipient)
-                                          for recipient in recipients}
+        self.message_channel_per_recipient_id = {
+            recipient.get_profile_id(): self.get_message_channel_by_recipient(recipient)
+            for recipient in self._recipients}
         self.provider_id_per_recipient = {recipient.get_profile_id(): self.get_message_provider_id(
-            message_channel_id=self.channel_ids_per_recipient[recipient.get_profile_id()], recipient=recipient)
-            for recipient in recipients}
+            message_channel=self.get_message_channel_by_recipient(recipient), recipient=recipient)
+            for recipient in self._recipients}
         # self.message_template_dict_by_campaign_id = self.get_message_template_dict_by_campaign_id(campaign_id=campaign_id)
 
         # Old code I might need later:
@@ -82,10 +84,10 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
         #     body = self.message_template.get_message_template_textblock_and_attributes(
         #         message_template_id=body, destination_profile_id=recipient.get_profile_id())
 
-    def get_id(self):
-        pass
+    def get_id(self) -> int or None:
+        return self.message_id
 
-    def get_msg_type(self, recipient: Recipient) -> MessageChannel:
+    def get_message_channel_by_recipient(self, recipient: Recipient) -> MessageChannel:
         # TODO: return msg_type (sms, email, whatsapp) based on hours, provider availability, msg length, etc.
         """TODO: make sure we can access:
         1. size of message
@@ -97,35 +99,40 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
         for message_channel in MessageChannel:
             if recipient.get_email_address() is not None:
                 return MessageChannel.EMAIL
-            elif len(self.get_body_text_after_template_processing(recipient, message_channel)) < SMS_MESSAGE_LENGTH:
+            elif len(self.get_body_text_after_template_processing(
+                    recipient=recipient, message_channel=message_channel)) < SMS_MESSAGE_LENGTH:
                 return MessageChannel.SMS
             else:
                 return MessageChannel.WHATSAPP
 
+    # TODO Let's create logger_table SQL to verify the decision taken by this method (like ...)
     def get_message_channel(self, recipient: Recipient) -> MessageChannel:
-        if recipient.get_profile_id() in self.channel_ids_per_recipient:
-            return self.channel_ids_per_recipient[recipient.get_profile_id()]
+        if recipient.get_profile_id() in self.message_channel_per_recipient_id:
+            message_channel = self.message_channel_per_recipient_id[recipient.get_profile_id()]
         else:
-            return self.get_msg_type(recipient)
+            message_channel = self.get_message_channel_by_recipient(recipient)
+        return message_channel
 
+    # TODO Let's create logger_table SQL to verify the decision taken by this method (like ...)
     @staticmethod
-    def get_message_provider_id(message_channel_id: MessageChannel, recipient: Recipient) -> int:
+    def get_message_provider_id(*, message_channel: MessageChannel, recipient: Recipient) -> int:
         """return message provider"""
         # TODO: implement the logic for the provider_id (remove +972, etc.)
-        if message_channel_id == MessageChannel.SMS and recipient.get_normizalied_phone().startswith("+972"):
+        if message_channel == MessageChannel.SMS and recipient.get_phone_number_full_normalized().startswith("972"):
             provider_id = AWS_SNS_SMS_MESSAGE_PROVIDER_ID
-        elif message_channel_id == MessageChannel.EMAIL:
+        elif message_channel == MessageChannel.EMAIL:
             provider_id = AWS_SES_EMAIL_MESSAGE_PROVIDER_ID
-        elif message_channel_id == MessageChannel.WHATSAPP:
+        elif message_channel == MessageChannel.WHATSAPP:
             # TODO: or vonage
             provider_id = INFORU_MESSAGE_PROVIDER_ID
         else:
-            raise Exception("Can't determine the Message Provider for message_channel_id=" + str(message_channel_id))
+            raise Exception("Can't determine the Message Provider for message_channel=" + str(message_channel))
 
         return provider_id
 
-    def get_subject_text_after_template_processing(self, recipient: Recipient,
-                                                   message_channel: MessageChannel = None) -> str:
+    @lru_cache
+    def get_subject_text_after_template_processing(
+            self, *, recipient: Recipient, message_channel: MessageChannel = None) -> str:
         seperator = ", "
         message_channel = message_channel or self.get_message_channel(recipient)
 
@@ -133,11 +140,15 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
         for block in self.get_profile_blocks(recipient.get_profile_id(), message_channel):
             preferred_language = block["preferred_language"]
             if block["subject_per_language"][preferred_language]:
-                subject += block["subject_per_language"][preferred_language] + seperator
+                subject_block = block["subject_per_language"][preferred_language]
+                if subject_block not in subject:
+                    subject += subject_block + seperator
+
         return subject
 
-    def get_body_text_after_template_processing(self, recipient: Recipient,
-                                                message_channel: MessageChannel = None) -> str:
+    @lru_cache
+    def get_body_text_after_template_processing(
+            self, *, recipient: Recipient, message_channel: MessageChannel = None) -> str:
         seperator = "\n"
         message_channel = message_channel or self.get_message_channel(recipient)
 
@@ -155,25 +166,30 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
 
         return body.rstrip(seperator)
 
-    # api_data To know if to API calls are actually the same and do caching.
-    # TODO sender_profile_id = None meaning we let the system to decide who is the sender_profile_id?
-    # TODO Do we need user_external as parameter?
-    def can_send(self, sender_profile_id: int = None, api_data: dict = None, outgoing_body: dict = None) -> bool:
-        sender_profile_id = sender_profile_id or self.sender_profile_id
-        if self._is_http_api:
-            return self.__can_send_direct(sender_profile_id, api_data=api_data)
-        else:
-            return self.__can_send_indirect(sender_profile_id, outgoing_body=outgoing_body)
+    def get_sender_profile_id(self, *, recipient: Recipient = None) -> int:
+        """determine the sender for each recipient based on (location, groups, campaign.is_dialog_workflow, campaign.default_sender_profile_id ...)"""
+        # TODO: implement the logic for the sender_profile_id
+        return self.__sender_profile_id  # or campaign_table.default_sender_profile_id
 
     # api_data To know if to API calls are actually the same and do caching.
-    def __can_send_direct(self, sender_profile_id: int, api_data: dict = None) -> bool:
+    # TODO Do we need user_external as parameter?
+    def can_send(self, *, sender_profile_id: int = None, api_data: dict = None, outgoing_body: dict = None,
+                 recipient: Recipient = None) -> bool:
+        sender_profile_id = sender_profile_id or self.get_sender_profile_id(recipient=recipient)
+        if self._is_http_api:
+            can_send = self.__can_send_direct(sender_profile_id=sender_profile_id, api_data=api_data)
+        else:
+            can_send = self.__can_send_indirect(sender_profile_id=sender_profile_id, outgoing_body=outgoing_body)
+        return can_send
+
+    # api_data To know if to API calls are actually the same and do caching.
+    def __can_send_direct(self, *, sender_profile_id: int, api_data: dict = None) -> bool:
         # TODO: implement sender_profile_id logic
         try:
-            # TODO: change try_to_call_api typing
             try_to_call_api_result = self.__direct.try_to_call_api(
                 user_external_id=self.__user_external_id,
                 api_type_id=self._api_type_id,
-                endpoint=self._endpoint,
+                endpoint_url=self._endpoint_url,
                 outgoing_body=api_data,  # data
                 outgoing_header=self._headers
             )
@@ -203,13 +219,13 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
             self.logger.exception(object=exception)
         return False
 
-    def __can_send_indirect(self, sender_profile_id: int = None, outgoing_body: dict = None) -> bool:
+    def __can_send_indirect(self, *, sender_profile_id: int = None, outgoing_body: dict = None) -> bool:
         # TODO: implement sender_profile_id logic
         http_status_code = None
         try:
             api_check, self.__api_call_id, arr = self.__indirect.before_call_api(
                 user_external_id=self.__user_external_id, api_type_id=self._api_type_id,
-                endpoint=self._endpoint,
+                endpoint_url=self._endpoint_url,
                 outgoing_header=self._headers,
                 outgoing_body=outgoing_body
             )
@@ -247,17 +263,17 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
         self.logger.info("http_status_code: " + str(http_status_code))
         return http_status_code == HTTPStatus.OK.value
 
-    def send(self, body: str = None, compound_message_dict: dict = None,
+    def send(self, *, body: str = None, compound_message_dict: dict = None,
              recipients: List[Recipient] = None, cc: List[Recipient] = None, bcc: List[Recipient] = None,
              scheduled_timestamp_start: datetime = None,
              scheduled_timestamp_end: datetime = None, **kwargs) -> list[int]:
         pass  # this is an abstract method, but we don't want to make this class abstract
 
-    def after_send_attempt(self, sender_profile_id: int = None, outgoing_body: dict = None,
-                           incoming_message: str = None, http_status_code: int = None,
-                           response_body: dict or json = None) -> None:
+    def after_send_attempt(self, *, sender_profile_id: int = None, outgoing_body: json = None,
+                           incoming_message: json = None, http_status_code: int = None,
+                           response_body: json or json = None, recipient: Recipient = None) -> None:
         # TODO: implement sender_profile_id logic
-        sender_profile_id = sender_profile_id or self.sender_profile_id
+        sender_profile_id = sender_profile_id or self.get_sender_profile_id(recipient=recipient)
         if self._is_http_api:
             self.__after_direct_send()
         else:
@@ -269,12 +285,12 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
     def display(self):
         print("MessageLocal: " + str(self.__dict__))
 
-    def __after_indirect_send(self, outgoing_body: dict, incoming_message: str,
-                              http_status_code: int, response_body: dict or json):
+    def __after_indirect_send(self, *, outgoing_body: json, incoming_message: json,
+                              http_status_code: int, response_body: json):
 
         self.__indirect.after_call_api(user_external_id=self.__user_external_id,
                                        api_type_id=self._api_type_id,
-                                       endpoint=self._endpoint,
+                                       endpoint_url=self._endpoint_url,
                                        outgoing_header=self._headers,
                                        outgoing_body=outgoing_body,
                                        incoming_message=incoming_message,
@@ -291,4 +307,5 @@ class MessageLocal(Item, CompoundMessage, metaclass=MetaLogger, object=object_me
         return self.importance
 
     def get_recipients(self) -> List[Recipient]:
-        return self.__recipients
+        # recipients may be modified by CompoundMessage
+        return self._recipients
